# Copyright 2020 Kapil Thangavelu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
from collections import defaultdict

import pytest
from pytest_terraform import hooks, tf, xdist


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    config.addinivalue_line("markers", "terraform: tests using terraform fixtures")

    cache_dir = config.getoption("dest_tf_plugin")
    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)
    tf.LazyPluginCacheDir.value = os.path.abspath(cache_dir)

    tf.LazyModuleDir.value = config.getoption("dest_tf_mod_dir") or config.getini(
        "terraform-mod-dir"
    )

    if tf.LazyReplay.value is None:
        tf.LazyReplay.value = config.getoption("dest_tf_replay")

    tf.LazyTfBin.value = config.getoption("dest_tf_binary") or shutil.which("terraform")
    if tf.LazyTfBin.value is None and not tf.LazyReplay.value:
        raise ValueError(
            "pytest-terraform requires terraform binary on PATH or "
            "specified with --tf-binary"
        )

    tf.PytestConfig.value = config

    if config.pluginmanager.hasplugin("xdist"):
        config.pluginmanager.register(xdist.XDistTerraform(config))
        tf.terraform.scope_class_map = d = defaultdict(
            lambda: xdist.ScopedTerraformFixture
        )
        d["function"] = tf.TerraformFixture


def pytest_addhooks(pluginmanager):
    """ Register pytest_terraform hooks """
    pluginmanager.add_hookspecs(hooks)


def pytest_addoption(parser):
    group = parser.getgroup("terraform")
    group.addoption(
        "--tf-binary",
        action="store",
        dest="dest_tf_binary",
        help=("Configure the path to the terraform binary. Default is to search PATH"),
    )
    group.addoption(
        "--tf-replay",
        action="store_true",
        dest="dest_tf_replay",
        help=("Use recorded resources instead of invoking terraform"),
    )
    group.addoption(
        "--tf-mod-dir",
        action="store",
        dest="dest_tf_mod_dir",
        help=("Configue the parent directory to look for terraform modules"),
    )
    group.addoption(
        "--tf-plugin-dir",
        action="store",
        dest="dest_tf_plugin",
        default=".tfcache",
        help=(
            "Use this directory for a terraform plugin cache "
            "Default is to use .tfcache"
        ),
    )

    parser.addini("terraform-mod-dir", "Parent Directory for terraform modules")



from .exceptions import InvalidTeardownMode


class TeardownOption:
    DEFAULT = None
    ON = "on"
    OFF = "off"
    IGNORE = "ignore"

    _options = (
        "on",
        "off",
        "default",
        "ignore",
    )

    def __init__(self, default="on"):
        self.set_default(default)

    def set_default(self, default):
        if default == "default" or default not in self._options:
            default = "on"
        self._default = default

    def resolve(self, option=None):
        if option is self.DEFAULT:
            return self._default

        option = option.lower()

        if option not in self._options:
            raise InvalidTeardownMode(
                "{} is not a valid option: {}".format(option, ",".join(self._options))
            )

        return option


teardown = TeardownOption()



# Copyright 2020 Kapil Thangavelu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from pytest_terraform import tf
from pytest_terraform.lock import lock_create, lock_delete


class ScopedTerraformFixture(tf.TerraformFixture):
    # specialized terraform fixture for use with
    # non function scopes, that is tracked for xdist
    # teardown, and locked around create/destroy

    state_dir = None
    wid = None
    _AutoTearDown = False

    def create(self, request, module_dir):
        if self.replay:
            super().create(request, module_dir)

        with lock_create(self.state_dir / self.name) as (success, result):
            if success:
                print(
                    "%s create %s - success: %s" % (self.wid, self.name, success),
                    file=sys.stderr,
                )
                tf_test_api = super(ScopedTerraformFixture, self).create(
                    request, module_dir
                )
                result.write(self.runner.work_dir.encode("utf8"))
                return tf_test_api
            return tf.TerraformTestApi.load(
                os.path.join(self.resolve_module_dir(), "tf_resources.json")
            )

    def tear_down(self):
        # print('%s %s fix teardown' % (self.wid, self.name), file=sys.stderr)
        with lock_delete(self.state_dir / self.name) as success:
            #  print('%s %s teardown state:%s' % (
            #        self.wid, self.name, success), file=sys.stderr)
            if not success:
                return
            work_dir = (self.state_dir / self.name).read_text("utf8")
            print(
                "%s teardown %s work-dir %s" % (self.wid, self.name, success),
                file=sys.stderr,
            )
            super(ScopedTerraformFixture)
            runner = self.get_runner(self.resolve_module_dir(), work_dir)
            runner.destroy()


class XDistTerraform(object):

    # Hooks
    # https://github.com/pytest-dev/pytest-xdist/blob/master/src/xdist/newhooks.py

    def __init__(self, config):
        self.config = config
        self.state_dir = None
        self.wid = None

        self.fixture_map = None  # only on worker nodes
        self.tracked_fixtures = set()  # only on worker nodes
        self.completed = set()  # only on worker nodes
        self.test_log = None  # read only fh on worker nodes, append fh on master
        self.activity = []

        if hasattr(self.config, "workerinput"):
            self.wid = self.config.workerinput["workerid"]
        else:
            self.wid = "master"

        if not getattr(self.config, "_tmpdirhandler", None):
            print("eeek %s" % self.wid)

        basetemp = self.config._tmpdirhandler.getbasetemp()
        if self.wid == "master":
            self.state_dir = basetemp.mkdir("terraform")
        else:
            self.state_dir = basetemp / ".." / "terraform"

        ScopedTerraformFixture.state_dir = self.state_dir
        ScopedTerraformFixture.wid = self.wid

        log_path = str(self.state_dir / "completed-log.txt")
        # print(log_path)
        if self.wid == "master":
            self.test_log_writer = open(log_path, mode="ab+", buffering=0)
        # make available to master for self unit testing
        self.test_log_reader = open(log_path, mode="r")

    def generate_fixture_map(self, items):
        fixture_map = {}
        for i in items:
            for f in i.fixturenames:
                if f in self.tracked_fixtures:
                    fixture_map.setdefault(f, set()).add(i.nodeid)
        return fixture_map

    # worker hooks
    def pytest_collection_modifyitems(self, session, config, items):
        """write out the collections of fixtures -> test ids

        in xdist this is only called from the workers
        """
        self.tracked_fixtures = {
            t.name
            for t in tf.terraform.get_fixtures()
            if isinstance(t, ScopedTerraformFixture)
        }
        self.fixture_map = self.generate_fixture_map(items)

    def pytest_runtest_teardown(self, item, nextitem):
        found = []
        for f in item.fixturenames:
            if f in self.tracked_fixtures:
                found.append(f)
        if not found:
            return
        # print(
        #    '%s worker teardown found: %s item used:%s tracked:%s' % (
        #    self.wid, found, item.fixturenames, self.tracked_fixtures), file=sys.stderr)

        completed = {n.strip() for n in self.test_log_reader.readlines()}
        self.completed.update(completed)
        self.completed.add(item.nodeid)
        # print("%s check teardown item:%s fixtures:%s completd:%s" % (
        #    self.wid, item.nodeid, found, self.completed), file=sys.stderr)
        for f in found:
            # print("%s check %s result %s" % (
            #       self.wid, f, self.completed.issuperset(self.fixture_map[f])),
            #       file=sys.stderr)
            # print('%s check result %s %s:%s' % (
            #       self.wid, completed, f, self.fixture_map[f]))
            if self.completed.issuperset(self.fixture_map[f]):
                print(
                    "%s execute test:%s teardown %s" % (self.wid, item.nodeid, f),
                    file=sys.stderr,
                )
                tf.terraform.get_fixture(f).tear_down()
                self.fixture_map.pop(f)

    def pytest_sessionfinish(self, exitstatus):
        if self.wid == "master":
            # print("master session finish", file=sys.stderr)
            return

        completed = {n.strip() for n in self.test_log_reader.readlines()}
        self.completed.update(completed)
        #        print("%s worker session down tracked:%d completed:%d %s" % (
        #            self.wid, len(self.tracked_fixtures),
        #            len(self.completed), completed), file=sys.stderr)
        #        print("%s worker map %s" % (
        #            self.wid, self.fixture_map), file=sys.stderr)

        remains = []
        for f in self.tracked_fixtures:
            if f not in self.fixture_map:
                continue
            if self.completed.issuperset(self.fixture_map[f]):
                print(
                    "%s worker session down cleanup %s" % (self.wid, f), file=sys.stderr
                )
                tf.terraform.get_fixture(f).tear_down()
            else:
                remains.append(str((f, self.fixture_map[f].difference(self.completed))))
        if remains:
            print("%s tf remains %s" % (self.wid, remains), file=sys.stderr)

    # master hooks
    def pytest_report_teststatus(self, report, config):
        # only called from master
        if self.wid != "master" or report.when != "call":
            return
        self.test_log_writer.write(("%s\n" % report.nodeid).encode("utf8"))
        self.test_log_writer.flush()
        os.fsync(self.test_log_writer.fileno())

    def pytest_configure_node(self, node):
        if not node.gateway.spec.popen:
            raise RuntimeError(
                "terraform plugin only compatible with xdist multi-process"
            )



class PytestTerraformError(Exception):
    """Pytest Terraform Exception Base Class"""


class TerraformCommandFailed(PytestTerraformError):
    """Terraform Command failed during execution"""


class InvalidOption(PytestTerraformError):
    """Invalid Option Error"""


class InvalidTeardownMode(InvalidOption):
    """Invalid Teardown Option Error"""


class InvalidState(PytestTerraformError):
    """Failure to load / parse state"""


class ModuleNotFound(ValueError):
    """module not found"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_terraform']

package_data = \
{'': ['*']}

install_requires = \
['jmespath>=0.10.0',
 'portalocker>=1.7.0',
 'pytest-xdist>=1.31.0',
 'pytest>=6.0']

entry_points = \
{'pytest11': ['terraform = pytest_terraform.plugin']}

setup_kwargs = {
    'name': 'pytest-terraform',
    'version': '0.6.1',
    'description': 'A pytest plugin for using terraform fixtures',
    'long_description': '# Introduction\n\n[![CI](https://github.com/cloud-custodian/pytest-terraform/workflows/CI/badge.svg?branch=master&event=push)](https://github.com/cloud-custodian/pytest-terraform/actions?query=branch%3Amaster)\n[![codecov](https://codecov.io/gh/cloud-custodian/pytest-terraform/branch/master/graph/badge.svg)](https://codecov.io/gh/cloud-custodian/pytest-terraform)\n\npytest_terraform is a pytest plugin that enables executing terraform\nto provision infrastructure in a unit/functional test as a fixture.\n\nThis plugin features uses a fixture factory pattern to enable paramterized\nconstruction of fixtures via decorators.\n\n## Usage\n\npytest_terraform provides a `terraform` decorator with the following parameters:\n\n| Argument             | Required? | Type    | Default      | Description |\n| -----                | :---:     | ---     | ---          | ---         |\n| `terraform_dir`      | yes       | String  |              | Terraform module (directory) to execute. |\n| `scope`              | no        | String  | `"function"` | [Pytest scope](https://docs.pytest.org/en/stable/fixture.html#scope-sharing-fixtures-across-classes-modules-packages-or-session) - should be one of: `function`, or `session`. Other scopes like  `class`, `module`, and `package` should work but have not been fully tested. |\n| `replay`             | no        | Boolean | `True`       | Use recorded resources instead of invoking terraform. See [Replay Support](#replay-support) for more details. |\n| `name`               | no        | String  | `None`       | Name used for the fixture. This defaults to the `terraform_dir` when `None` is supplied. |\n| `teardown`           | no        | String  | `"default"`  | Configure which teardown mode is used for terraform resources. See [Teardown Options](#teardown-options) for more details. |\n\n### Example\n\n```python\nfrom boto3 import Session\nfrom pytest_terraform import terraform\n\n\n# We use the terraform decorator to create a fixture with the name of\n# the terraform module.\n#\n# The test function will be invoked after the terraform module is provisioned\n# with the results of the provisioning.\n#\n# The module `aws_sqs` will be searched for in several directories, the test\n# file directory, a sub directory `terraform`.\n#\n# This fixture specifies a session scope and will be run once per test run.\n#\n@terraform(\'aws_sqs\', scope=\'session\')\ndef test_sqs(aws_sqs):\n    # A test is passed a terraform resources class containing content from\n    # the terraform state file.\n    #\n    # Note the state file contents may vary across terraform versions.\n    #\n    # We can access nested datastructures with a jmespath expression.\n    assert aws_sqs["aws_sqs_queue.test_queue.tags"] == {\n        "Environment": "production"\n    }\n   queue_url = aws_sqs[\'test_queue.queue_url\']\n   print(queue_url)\n\n\ndef test_sqs_deliver(aws_sqs):\n   # Once a fixture has been defined with a decorator\n   # it can be reused in the same module by name, with provisioning\n   # respecting scopes.\n   #\n   sqs = Session().client(\'sqs\')\n   sqs.send_message(\n       QueueUrl=aws_sqs[\'test_queue.queue_url\'],\n       MessageBody=b"123")\n\n\n@terraform(\'aws_sqs\')\ndef test_sqs_dlq(aws_sqs):\n   # the fixture can also referenced again via decorator, if redefined\n   # with decorator the fixture parameters much match (ie same session scope).\n\n   # Module outputs are available as a separate mapping.\n   aws_sqs.outputs[\'QueueUrl\']\n```\n\n*Note* the fixture name should match the terraform module name\n\n*Note* The terraform state file is considered an internal\nimplementation detail of terraform, not per se a stable public interface\nacross versions.\n\n## Marks\n\nAll tests using terraform fixtures have a `terraform` mark applied so\nthey can be run/selected via the command line ie.\n\n```shell\npytest -k terraform tests/\n```\n\nto run all terraform tests only. See pytest mark documentation for\nadditional details, https://docs.pytest.org/en/stable/example/markers.html#mark-examples\n\n\n## Options\n\nYou can provide the path to the terraform binary else its auto discovered\n```shell\n--tf-binary=$HOME/bin/terraform\n```\n\nTo avoid repeated downloading of plugins a plugin cache dir is utilized\nby default this is `.tfcache` in the current working directory.\n```shell\n--tf-plugin-dir=$HOME/.cache/tfcache\n```\n\nTerraform modules referenced by fixtures are looked up in a few different\nlocations, directly in the same directory as the test module, in a subdir\nnamed terraform, and in a sibling directory named terraform. An explicit\ndirectory can be given which will be looked at first for all modules.\n\n```shell\n--tf-mod-dir=terraform\n```\n\nThis plugin also supports flight recording (see next section)\n```shell\n--tf-replay=[record|replay|disable]\n```\n\n### Teardown Options\n\n`pytest_terraform` supports three different teardown modes for the terraform decorator.\nThe default, `pytest_terraform.teardown.ON` will always attempt to teardown any and all modules via `terraform destory`.\nIf for any reason destroy fails it will raise an exception to alert the test runner.\nThe next mode, `pytest_terraform.teardown.IGNORE`, will invoke `terraform destroy` as with `teardown.ON` but will ignore any failures.\nThis mode is particularly help if your test function performs destructive actions against any objects created by the terraform module.\nThe final option is `pytest_terraform.teardown.OFF` which will remove the teardown method register all together.\nThis should generally only be used in very specific situations and is considered an edge case.\n\nThere is a special `pytest_terraform.teardown.DEFAULT` which is what the `teardown` parameter actually defaults to.\n\nTeardown options are available, for convenience, on the terraform decorator.\nFor example, set teardown to ignore:\n\n```python\nfrom pytest_terraform import terraform\n\n\n@terraform(\'aws_sqs\', teardown=terraform.TEARDOWN_IGNORE)\ndef test_sqs(aws_sqs):\n    assert aws_sqs["aws_sqs_queue.test_queue.tags"] == {\n        "Environment": "production"\n    }\n   queue_url = aws_sqs[\'test_queue.queue_url\']\n   print(queue_url)\n```\n\n## Hooks\n\npytest_terraform provides hooks via the pytest hook implementation.\nHooks should be added in the `conftest.py` file.\n\n### `pytest_terraform_modify_state`\n\nThis hook is executed after state has been captured from terraform apply and before writing to disk.\nThis hook does not modify state that\'s passed to the function under test.\nThe state is passed as the kwarg `tfstate` which is a `TerraformStateJson` UserString class with the following methods and properties:\n\n- `TerraformStateJson.dict` - The deserialized state as a dict\n- `TerraformStateJson.update(state: str)` - Replace the serialized state with a new state string\n- `TerraformStateJson.update_dict(state: dict)` - Replace the serialized state from a dictionary\n\n#### Example\n\n```python\ndef pytest_terraform_modify_state(tfstate):\n    print(str(tfstate))\n```\n\n#### Example AWS Account scrub\n\n```python\nimport re\n\ndef pytest_terraform_modify_state(tfstate):\n    """ Replace potential AWS account numbers with \'REDACTED\' """\n    tfstate.update(re.sub(r\'([0-9]+){12}\', \'REDACTED\', str(tfstate)))\n```\n\n## Flight Recording\n\nThe usage/philosophy of this plugin is based on using flight recording\nfor unit tests against cloud infrastructure. In flight recording rather\nthan mocking or stubbing infrastructure, actual resources are created\nand interacted with with responses recorded, with those responses\nsubsequently replayed for fast test execution. Beyond the fidelity\noffered, this also enables these tests to be executed/re-recorded against\nlive infrastructure for additional functional/release testing.\n\nhttps://cloudcustodian.io/docs/developer/tests.html#creating-cloud-resources-with-terraform\n\n### Replay Support\n\nBy default fixtures will save a `tf_resources.json` back to the module\ndirectory, that will be used when in replay mode.\n\nReplay can be configured by passing --tf-replay on the cli or via pytest config file.\n\n### Recording\n\nPassing the fixture parameter `replay` can control the replay behavior on an individual\ntest. The default is to operate in recording mode.\n\n@terraform(\'file_example\', replay=False)\ndef test_file_example(file_example):\n    assert file_example[\'local_file.bar.content\'] == \'bar!\'\n\n\n## XDist Compatibility\n\npytest_terraform supports pytest-xdist in multi-process (not distributed)\nmode.\n\nWhen run with python-xdist, pytest_terraform treats all non functional\nscopes as per test run fixtures across all workers, honoring their\noriginal scope lifecycle but with global semantics, instead of once\nper worker (xdist default).\n\nTo enable this the plugin does multi-process coodination using lock\nfiles, a test execution log, and a dependency mapping of fixtures\nto tests. Any worker can execute a module teardown when its done executing\nthe last test that depends on a given fixture. All provisioning and\nteardown are guarded by atomic file locks in the pytest execution\'s temp\ndirectory.\n\n### Root module references\n\n`terraform_remote_state` can be used to introduce a dependency between\na scoped root modules on an individual test, note we are not\nattempting to support same scope inter fixture dependencies as that\nimposes additional scheduling constraints outside of pytest native\ncapabilities. The higher scoped root module (ie session or module scoped)\nwill need to have output variables to enable this consumption.\n',
    'author': 'Kapil Thangavelu',
    'author_email': 'kapilt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cloud-custodian/pytest-terraform',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020 Kapil Thangavelu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import subprocess
import sys
from collections import UserString, defaultdict
from typing import Any, Dict, Optional, Tuple, Union

import jmespath
import pytest
from py.path import local

from .exceptions import InvalidState, ModuleNotFound, TerraformCommandFailed
from .options import teardown as td


class TerraformRunner(object):

    command_templates = {
        "init": "init {input} {color} {plugin_dir}",
        "apply": "apply {input} {color} {state} {approve} {plan}",
        "plan": "plan {input} {color} {state} {output}",
        "destroy": "destroy {input} {color} {state} {approve}",
        "show": "show {color} -json {state_path}",
    }

    template_defaults = {
        "input": "-input=false",
        "color": "-no-color",
        "approve": "-auto-approve",
    }

    debug = False

    def __init__(
        self,
        work_dir,
        state_path=None,
        module_dir=None,
        plugin_cache=None,
        stream_output=None,
        tf_bin=None,
    ):

        self.work_dir = work_dir
        self.module_dir = module_dir
        # use parent dir of work/data dir to avoid
        # https://github.com/hashicorp/terraform/issues/22999
        self.state_path = state_path or os.path.join(work_dir, "..", "terraform.tfstate")
        self.stream_output = stream_output
        self.plugin_cache = plugin_cache or ""
        self.tf_bin = tf_bin

    def apply(self, plan=True):
        """run terraform apply"""
        if plan is True:
            plan_path = os.path.join(self.work_dir, "tfplan")
            self.plan(plan_path)
            apply_args = self._get_cmd_args("apply", plan=plan_path)
        elif plan:
            apply_args = self._get_cmd_args("apply", plan="")
        self._run_cmd(apply_args)
        return TerraformState.from_file(self.state_path, self)

    def plan(self, output=""):
        output = output and "-out=%s" % output or ""
        self._run_cmd(self._get_cmd_args("plan", output=output))

    def init(self):
        self._run_cmd(self._get_cmd_args("init", plugin_dir=""))

    def destroy(self):
        self._run_cmd(self._get_cmd_args("destroy"))

    def show(self):
        return json.loads(
            self._run_cmd(
                self._get_cmd_args("show", state_path=self.state_path), output=True
            ).decode("utf8")
        )

    def _get_cmd_args(self, cmd_name, tf_bin=None, env=None, **kw):
        tf_bin = tf_bin and tf_bin or self.tf_bin
        kw.update(self.template_defaults)
        kw["state"] = self.state_path and "-state=%s" % self.state_path or ""
        return [tf_bin] + list(
            filter(None, self.command_templates[cmd_name].format(**kw).split(" "))
        )

    def _run_cmd(self, args, output=False):
        env = dict(os.environ)
        tf_env = {}
        if LazyPluginCacheDir.resolve():
            tf_env["TF_PLUGIN_CACHE_DIR"] = LazyPluginCacheDir.resolve()
        tf_env["TF_IN_AUTOMATION"] = "yes"
        if self.module_dir:
            tf_env["TF_DATA_DIR"] = self.work_dir
        cwd = self.module_dir or self.work_dir
        env.update(tf_env)
        print("run cmd", args, tf_env, cwd, file=sys.stderr)
        run_cmd = subprocess.check_call
        if output:
            run_cmd = subprocess.check_output
        return run_cmd(args, cwd=cwd, stderr=subprocess.STDOUT, env=env)


class TerraformStateJson(UserString):
    @classmethod
    def from_dict(cls, state: Dict[str, Any]):
        """create TerraformStateJson from dictionary"""
        s = cls("")
        s.update_dict(state)
        return s

    def update(self, state: str):
        """update TerraformStateJson object with new data"""
        if not isinstance(state, str):
            raise ValueError(f"{state} is not a string")

        self.data = str(state)

    def update_dict(self, state: Dict[str, Any]):
        """update TerraformStateJson from a dict"""
        self.update(json.dumps(state, indent=4))

    @property
    def dict(self):
        """return the TerraformStateJson as a dict"""
        return json.loads(self.data)

    @dict.setter
    def dict(self, data: Dict[str, Any]):
        """update TerraformStateJson from a dict"""
        try:
            self.update_dict(data)
        except (ValueError, TypeError):
            raise ValueError("Not a serializable object")


class TerraformState(object):
    """Abstraction over a terrafrom state file with helpers.

    resources dict contains a minimal representation of a terraform
    state file with enough identity information to query a resource
    from the api.

    resources dict is a nested data structure corresponding to
       resource_type -> resource name -> resource attributes.

    by default all resources will have an 'id' attribute, additional
    attributes which contain the key 'name' will also be present.
    """

    def __init__(self, resources, outputs, runner=None):
        self._runner = runner
        self.outputs = outputs
        self.resources = resources

    @property
    def work_dir(self):
        if self._runner:
            return self._runner.work_dir

    @property
    def terraform(self):
        if self._runner:
            return self._runner

    def __getitem__(self, k):
        v = self.get(k)
        if v is None:
            raise KeyError(k)
        return v

    def get(self, k, default=None):
        """accessor to resource attributes.

        supports a few shortcuts for ease of use.

        key can be a jmespath expression in which case the evaluation
        is returned.

        if key is a unique resource name, then its data is returned, if
        the data is a singleton key dictionary with 'id', then just then
        the string value of 'id' is returned.
        """
        if "." in k:
            return jmespath.search(k, self.resources)
        found = False
        for rtype in self.resources:
            for rname in self.resources[rtype]:
                if rname == k:
                    assert found is False, "Ambigious resource name %s" % k
                    found = self.resources[rtype][rname]
        if found:
            if len(found) == 1:
                return found["id"]
            return found
        return default

    @classmethod
    def from_file(cls, path: str, runner=None):
        """create TerraformState from a file

        File can either be a Terraform Plan state, or a recorded
        pytest-terraform state
        """
        if not os.path.isfile(path):
            raise InvalidState("{} could not be located".format(path))

        with open(path) as fh:
            state = fh.read()

        return cls.from_string(state, runner)

    @classmethod
    def from_string(cls, state: Union[TerraformStateJson, str], runner=None):
        """create TerraformState from string

        State string can be a bytestring or a TerraformStateJson
        string object
        """
        resources, outputs = cls.parse_state(state)
        return cls(resources, outputs, runner)

    def update(self, state: Union[TerraformStateJson, str]):
        """update TerraformState values"""
        resources, outputs = self.parse_state(state)
        self.resources = resources
        self.outputs = outputs

    @staticmethod
    def parse_state(
        state: Union[TerraformStateJson, str]
    ) -> Tuple[Dict[str, any], Dict[str, Any]]:
        """extract resources and outputs from state

        where state is one of the following:
        * Terraform state output as a string
        * Recorded pytest-terraform state
        * TerraformStateJson object
        """
        if isinstance(state, TerraformStateJson):
            data = state.dict
        else:
            data = json.loads(state)

        if "pytest-terraform" in data:
            return (data["resources"], data["outputs"])

        resources = {}
        outputs = {}

        for r in data.get("resources", ()):
            rmap = resources.setdefault(r["type"], {})
            rmap[r["name"]] = dict(r["instances"][0]["attributes"])

        outputs = data.get("outputs", {})
        for m in data.get("modules", ()):
            for k, r in m.get("resources", {}).items():
                if k.startswith("data"):
                    continue
                module, rname = k.split(".", 1)
                rmap = resources.setdefault(module, {})
                rattrs = {"id": r["primary"]["id"]}
                for kattr, vattr in r["primary"]["attributes"].items():
                    if "name" in kattr and vattr != rattrs["id"]:
                        rattrs[kattr] = vattr
                rmap[rname] = rattrs

        return (resources, outputs)

    def export(self):
        """export state as a TerraformStateJson UserString"""

        state = {
            "pytest-terraform": 1,
            "outputs": self.outputs,
            "resources": self.resources,
        }

        return TerraformStateJson.from_dict(state)

    def save(self, state_path: Optional[str] = None) -> Optional[TerraformStateJson]:
        """export state to a file"""

        output = self.export()
        if not state_path:
            return output

        with open(state_path, "w") as fh:
            fh.write(str(output))


class TerraformTestApi(TerraformState):
    """public api to tests as fixture value."""


class PlaceHolderValue(object):
    """Lazy / Late resolved named values.

    many of our instantiations are at module import time, to support
    runtime configuration from cli/ini options we utilize a lazy
    loaded value set which is configured for final values via hooks
    (early, post conf, pre collection).
    """

    def __init__(self, name):
        self.name = name
        self.value = None

    def resolve(self, default=None):
        if not self.value and default:
            raise ValueError("PlaceHolderValue %s not resolved" % self.name)
        return self.value or default


LazyReplay = PlaceHolderValue("tf_replay")
LazyModuleDir = PlaceHolderValue("module_dir")
LazyPluginCacheDir = PlaceHolderValue("plugin_cache")
LazyTfBin = PlaceHolderValue("tf_bin_path")
PytestConfig = PlaceHolderValue("pytestconfig")


class TerraformFixture(object):
    def __init__(
        self,
        tf_bin,
        plugin_cache,
        scope,
        tf_root_module,
        test_dir,
        replay,
        teardown,
        pytest_config,
    ):
        self.tf_bin = tf_bin
        self.tf_root_module = tf_root_module
        self.test_dir = test_dir
        self.scope = scope
        self.replay = replay
        self.runner = None
        self.teardown_config = td.resolve(teardown)
        self.config = pytest_config

    @property
    def name(self):
        return "%s" % self.tf_root_module

    __name__ = name

    def resolve_module_dir(self):
        candidates = [
            self.test_dir.join(self.tf_root_module),
            self.test_dir.join("terraform", self.tf_root_module),
            self.test_dir.dirpath().join(self.tf_root_module),
            self.test_dir.dirpath().join("terraform", self.tf_root_module),
        ]
        if LazyModuleDir.resolve():
            candidates.insert(0, local(LazyModuleDir.resolve()).join(self.tf_root_module))
        for candidate in candidates:
            if not candidate.check(exists=1, dir=1):
                continue
            return candidate
        raise ModuleNotFound(self.tf_root_module)

    def get_runner(self, module_dir, work_dir):
        return TerraformRunner(
            str(work_dir),
            module_dir=module_dir,
            plugin_cache=LazyPluginCacheDir.resolve(),
            tf_bin=LazyTfBin.resolve(),
        )

    def __call__(self, request, tmpdir_factory, worker_id):
        module_dir = self.resolve_module_dir()
        if self.replay:
            replay_resources = os.path.join(module_dir, "tf_resources.json")
            if not os.path.exists(replay_resources):
                raise ValueError(
                    "Replay resources don't exist for %s" % self.tf_root_module
                )
            return TerraformTestApi.from_file(
                os.path.join(module_dir, "tf_resources.json")
            )
        work_dir = tmpdir_factory.mktemp(self.tf_root_module, numbered=True).join("work")
        self.runner = self.get_runner(module_dir, work_dir)
        return self.create(request, module_dir)

    def create(self, request, module_dir):
        print("tf create %s" % self.tf_root_module, file=sys.stderr)
        self.runner.init()
        if self.teardown_config != td.OFF:
            request.addfinalizer(self.tear_down)
        try:
            state = self.runner.apply()
            state_json = state.export()
            test_api = TerraformTestApi.from_string(state_json, self.runner)

            self.config.hook.pytest_terraform_modify_state(tfstate=state_json)

            state.update(state_json)
            state.save(module_dir.join("tf_resources.json"))

            return test_api
        except Exception:
            raise

    def tear_down(self):
        # config behavor on runner
        print("tf teardown %s" % self.tf_root_module, file=sys.stderr)
        try:
            self.runner.destroy()
        except subprocess.CalledProcessError as e:
            if self.teardown_config == td.IGNORE:
                return
            raise TerraformCommandFailed from e


class FixtureDecoratorFactory(object):
    """Generate fixture decorators on the fly."""

    scope_class_map = defaultdict(lambda: TerraformFixture)

    # Make accessing teardown options easier for users
    TEARDOWN_IGNORE = td.IGNORE
    TEARDOWN_OFF = td.OFF
    TEARDOWN_ON = td.ON

    def __init__(self):
        self._fixtures = []

    def get_fixtures(self):
        return list(self._fixtures)

    def get_fixture(self, name):
        for f in self._fixtures:
            if f.name == name:
                return f
        raise KeyError(name)

    def __call__(
        self,
        terraform_dir,
        scope="function",
        replay=None,
        name=None,
        teardown=td.DEFAULT,
    ):
        # We have to hook into where fixture discovery will find
        # our fixtures, the easiest option is to store on the module that
        # originated the call, all test modules get scanned for
        # fixtures. The alternative is to try and keep a set and
        # store. this particular setup is to support decorator usage.
        # ie. its gross on one hand and very pratical for consumers
        # on the other.
        f = sys._getframe(1)
        name = name or terraform_dir
        test_dir = local(_frame_path(f)).dirpath()
        if replay is None:
            replay = LazyReplay.resolve()
        found = None
        for tf in self._fixtures:
            if tf.tf_root_module == terraform_dir:
                assert scope == tf.scope, (
                    "Same tf module:%s used at different scopes"
                ) % (terraform_dir)
                found = tf
        if found:
            return self.nonce_decorator
        tclass = self.scope_class_map[scope]
        tfix = tclass(
            LazyTfBin,
            LazyPluginCacheDir,
            scope,
            terraform_dir,
            test_dir,
            replay,
            teardown,
            PytestConfig.resolve(),
        )
        self._fixtures.append(tfix)
        marker = pytest.fixture(scope=scope, name=terraform_dir)
        f.f_locals[name] = marker(tfix)
        return self.nonce_decorator

    @staticmethod
    def nonce_decorator(func):
        pytest.mark.terraform(func)
        return func


def _frame_path(f):
    start = f
    while f:
        if "__file__" in f.f_locals:
            return f.f_locals["__file__"]
        f = f.f_back
    raise RuntimeError("frame path not found %s" % start)


terraform = FixtureDecoratorFactory()



# Copyright 2020 Kapil Thangavelu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import portalocker
from py.path import local

PollInterval = 5
LockTimeout = 300


@contextlib.contextmanager
def lock_create(file_path, timeout=LockTimeout, interval=PollInterval):
    """Context manager for file creation with file locking

    return a tuple either
        - (false, file_content)
        - (true, file_handle)
    """
    fp = local(file_path)
    if fp.exists():
        yield False, fp.read_text("utf8")
        return
    with portalocker.Lock(
        str(fp.dirpath() / (fp.basename + ".lock")),
        timeout=timeout,
        check_interval=interval,
    ):
        if fp.exists():
            yield False, fp.read_text("utf8")
            return
        with portalocker.open_atomic(str(fp)) as fh:
            yield True, fh
            fh.flush()


@contextlib.contextmanager
def lock_delete(file_path, timeout=LockTimeout, interval=PollInterval):
    """Context manager for file delete with file locking

    Returns boolean
      - False if file didn't exist
      - True if file will be deleted
    """
    pointer = local(file_path)
    if not pointer.exists():
        yield False
        return
    with portalocker.Lock(
        str(pointer.dirpath() / (pointer.basename + ".lock")),
        timeout=timeout,
        check_interval=interval,
    ):
        if not pointer.exists():
            yield False
            return
        yield True
        pointer.remove()
